﻿using System;
using System.Collections.Generic;
using System.Data.Entity;
using System.Data.Entity.Core.Metadata.Edm;
using System.Data.Entity.Core.Objects;
using System.Data.Entity.Infrastructure;
using System.Linq;
using System.Reflection;

namespace fgxavi.Common.Business.Repositories
{
    public static class DbContextExtensions
    {
        /// <summary>
        /// Obté un diccionari amb totes les entitats del context de la base de dades.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <returns>Diccionari amb totes les entitats del context de la base de dades.</returns>
        public static IDictionary<Type, String> GetTables(this DbContext ctx)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            IEnumerable<EntityType> entities = octx.MetadataWorkspace.GetItemCollection(DataSpace.OSpace).GetItems<EntityType>().ToList();
            return (entities.ToDictionary(x => Type.GetType(x.FullName), x => GetTableName(ctx, Type.GetType(x.FullName))));
        }

        /// <summary>
        /// Obté el nom de la taula de l'entitat.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Nom de la taula de l'entitat.</returns>
        public static String GetTableName(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntitySetBase et = octx.MetadataWorkspace.GetItemCollection(DataSpace.SSpace)
                .GetItems<EntityContainer>()
                .Single()
                .BaseEntitySets
                .Where(x => x.Name == entityType.Name)
                .Single();
            String tableName = String.Concat(et.MetadataProperties["Schema"].Value, ".", et.MetadataProperties["Table"].Value);
            return (tableName);
        }

        /// <summary>
        /// Obté les propietats de navegació de l'entitat amb multiplicitat 1-n.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Propietats de navegació de l'entitat amb multiplicitat 1-n.</returns>
        public static IEnumerable<PropertyInfo> OneToMany(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntityType et = octx.MetadataWorkspace.GetItems(DataSpace.OSpace).Where(x => x.BuiltInTypeKind == BuiltInTypeKind.EntityType).OfType<EntityType>().Where(x => x.Name == entityType.Name).Single();
            return (et.NavigationProperties.Where(x => x.FromEndMember.RelationshipMultiplicity == RelationshipMultiplicity.One && x.ToEndMember.RelationshipMultiplicity == RelationshipMultiplicity.Many).Select(x => entityType.GetProperty(x.Name, BindingFlags.Public | BindingFlags.Instance | BindingFlags.GetProperty | BindingFlags.SetProperty)).ToList());
        }

        /// <summary>
        /// Obté les propietats de navegació de l'entitat amb multiplicitat 1-1.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Propietats de navegació de l'entitat amb multiplicitat 1-1.</returns>
        public static IEnumerable<PropertyInfo> OneToOne(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntityType et = octx.MetadataWorkspace.GetItems(DataSpace.OSpace).Where(x => x.BuiltInTypeKind == BuiltInTypeKind.EntityType).OfType<EntityType>().Where(x => x.Name == entityType.Name).Single();
            return (et.NavigationProperties.Where(x => (x.FromEndMember.RelationshipMultiplicity == RelationshipMultiplicity.One || x.FromEndMember.RelationshipMultiplicity == RelationshipMultiplicity.ZeroOrOne) && (x.ToEndMember.RelationshipMultiplicity == RelationshipMultiplicity.One || x.ToEndMember.RelationshipMultiplicity == RelationshipMultiplicity.ZeroOrOne)).Select(x => entityType.GetProperty(x.Name, BindingFlags.Public | BindingFlags.Instance | BindingFlags.GetProperty | BindingFlags.SetProperty)).ToList());
        }

        /// <summary>
        /// Obté les propietats de navegació de l'entitat amb multiplicitat n-1.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Propietats de navegació de l'entitat amb multiplicitat n-1.</returns>
        public static IEnumerable<PropertyInfo> ManyToOne(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntityType et = octx.MetadataWorkspace.GetItems(DataSpace.OSpace).Where(x => x.BuiltInTypeKind == BuiltInTypeKind.EntityType).OfType<EntityType>().Where(x => x.Name == entityType.Name).Single();
            return (et.NavigationProperties.Where(x => x.FromEndMember.RelationshipMultiplicity == RelationshipMultiplicity.Many && x.ToEndMember.RelationshipMultiplicity == RelationshipMultiplicity.One).Select(x => entityType.GetProperty(x.Name, BindingFlags.Public | BindingFlags.Instance | BindingFlags.GetProperty | BindingFlags.SetProperty)).ToList());
        }

        /// <summary>
        /// Obté les propietats de navegació de l'entitat amb multiplicitat n-n.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Propietats de navegació de l'entitat amb multiplicitat n-n.</returns>
        public static IEnumerable<PropertyInfo> ManyToMany(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntityType et = octx.MetadataWorkspace.GetItems(DataSpace.OSpace).Where(x => x.BuiltInTypeKind == BuiltInTypeKind.EntityType).OfType<EntityType>().Where(x => x.Name == entityType.Name).Single();
            return (et.NavigationProperties.Where(x => x.FromEndMember.RelationshipMultiplicity == RelationshipMultiplicity.Many && x.ToEndMember.RelationshipMultiplicity == RelationshipMultiplicity.Many).Select(x => entityType.GetProperty(x.Name, BindingFlags.Public | BindingFlags.Instance | BindingFlags.GetProperty | BindingFlags.SetProperty)).ToList());
        }

        /// <summary>
        /// Obté les propietats de les claus primàries de l'entitat.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Propietats de les claus primàries de l'entitat.</returns>
        public static IEnumerable<PropertyInfo> GetIdProperties(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntityType et = octx.MetadataWorkspace.GetItems(DataSpace.OSpace).Where(x => x.BuiltInTypeKind == BuiltInTypeKind.EntityType).OfType<EntityType>().Where(x => x.Name == entityType.Name).Single();
            return (et.KeyMembers.Select(x => entityType.GetProperty(x.Name)).ToList());
        }

        /// <summary>
        /// Obté les propietats de navegació de l'entitat.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Propietats de navegació de l'entitat.</returns>
        public static IEnumerable<PropertyInfo> GetNavigationProperties(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntityType et = octx.MetadataWorkspace.GetItems(DataSpace.OSpace).Where(x => x.BuiltInTypeKind == BuiltInTypeKind.EntityType).OfType<EntityType>().Where(x => x.Name == entityType.Name).Single();
            return (et.NavigationProperties.Select(x => entityType.GetProperty(x.Name)).ToList());
        }

        /// <summary>
        /// Obté un diccionari amb les claus primàries de l'entitat.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Diccionari amb les claus primàries de l'entitat.</returns>
        public static IDictionary<String, PropertyInfo> GetTableKeyColumns(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntityType objectEntityType = octx.MetadataWorkspace.GetItems(DataSpace.OSpace).Where(x => x.BuiltInTypeKind == BuiltInTypeKind.EntityType).OfType<EntityType>().Where(x => x.Name == entityType.Name).Single();
            return (objectEntityType.KeyMembers.Select((elm, index) => new { elm.Name, Property = entityType.GetProperty((objectEntityType.MetadataProperties["Members"].Value as IEnumerable<EdmMember>).ElementAt(index).Name) }).ToDictionary(x => x.Name, x => x.Property));
        }

        /// <summary>
        /// Obté un diccionari amb les propietats de l'entitat.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Diccionari amb les propietats de l'entitat.</returns>
        public static IDictionary<String, PropertyInfo> GetTableColumns(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntityType objectEntityType = octx.MetadataWorkspace.GetItems(DataSpace.OSpace).Where(x => x.BuiltInTypeKind == BuiltInTypeKind.EntityType).OfType<EntityType>().Where(x => x.Name == entityType.Name).Single();
            return (objectEntityType.Properties.Select((elm, index) => new { elm.Name, Property = entityType.GetProperty(objectEntityType.Members[index].Name) }).ToDictionary(x => x.Name, x => x.Property));
        }

        /// <summary>
        /// Obté un diccionari amb les propietats de navegació de l'entitat.
        /// </summary>
        /// <param name="ctx">Context de la base de dades.</param>
        /// <param name="entityType">Tipus de l'entitat.</param>
        /// <returns>Diccionari amb les propietats de navegació de l'entitat.</returns>
        public static IDictionary<String, PropertyInfo> GetTableNavigationColumns(this DbContext ctx, Type entityType)
        {
            ObjectContext octx = (ctx as IObjectContextAdapter).ObjectContext;
            EntityType objectEntityType = octx.MetadataWorkspace.GetItems(DataSpace.OSpace).Where(x => x.BuiltInTypeKind == BuiltInTypeKind.EntityType).OfType<EntityType>().Where(x => x.Name == entityType.Name).Single();
            return (objectEntityType.NavigationProperties.Select((elm, index) => new { elm.Name, Property = entityType.GetProperty(objectEntityType.Members[index].Name) }).ToDictionary(x => x.Name, x => x.Property));
        }
    }
}
