﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Threading.Tasks;

namespace fgxavi.Common.Business.Repositories
{
    public interface IRepository<TEntity, TKey>
        where TEntity : class
    {
        Func<TEntity, TKey> KeySelector { get; set; }
        IDictionary<string, PropertyInfo> Keys { get; }
        IDictionary<string, PropertyInfo> Fields { get; }
        IDictionary<string, PropertyInfo> Navigations { get; }
        /// <summary>
        /// Inicialitza la instància amb una funció que obté la clau primària d'una entitat.
        /// </summary>
        /// <remarks>Cal executar aquest mètode tan bon punt es crei la instància perquè el mètode GetKey retorni el valor correcte, en cas de no executar-se el mètode GetKey retornarà el valor per defecte del tipus de la clau primària.</remarks>
        void Initialize(Func<TEntity, TKey> keySelector);
        /// <summary>
        /// Obté una consulta IQueryable de totes les entitats de tipus TEntity de la base de dades.
        /// </summary>
        /// <remarks>No es realitza cap consulta a la base de dades fins que no s'executa la consulta IQueryable convertint-se en una enumeració IEnumerable.</remarks>
        /// <returns>Consulta IQueryable de totes les entitats de tipus TEntity.</returns>
        IQueryable<TEntity> GetAll();
        /// <summary>
        /// Obté l'entitat de tipus TEntity que té com a clau primària el tipus TKey facilitat.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <returns>Entitat coincident.</returns>
        TEntity GetByKey(TKey key);
        /// <summary>
        /// Obté l'entitat de tipus TEntity que té com a clau primària el tipus TKey facilitat.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <returns>Entitat coincident.</returns>
        Task<TEntity> GetByKeyAsync(TKey key);
        /// <summary>
        /// Obté la clau primària de tipus TKey de l'entitat.
        /// </summary>
        /// <remarks>Cal inicialitzar la instància de la classe amb el mètode Initialize per definir la funció que retorna la clau primària d'una entitat. En cas de no definir-se la funció es retorna el valor per defecte del tipus de la clau primària.</remarks>
        /// <param name="entity">Entitat per a obtenir la clau primària.</param>
        /// <returns>Clau primària de l'entitat.</returns>
        TKey GetKey(TEntity entity);
        /// <summary>
        /// Crea una nova entitat a la base de dades.
        /// </summary>
        /// <param name="entity">Entitat per afegir al context.</param>
        /// <returns>Entitat creada.</returns>
        TEntity Create(TEntity entity);
        /// <summary>
        /// Crea una nova entitat a la base de dades.
        /// </summary>
        /// <param name="entity">Entitat per afegir al context.</param>
        /// <returns>Entitat creada.</returns>
        /// <remarks>Versió asíncrona.</remarks>
        Task<TEntity> CreateAsync(TEntity entity);
        /// <summary>
        /// Actualitza l'entitat modificada a la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="update">Entitat modificada per actualitzar al context.</param>
        TEntity Update(TKey key, TEntity update);
        /// <summary>
        /// Actualitza l'entitat modificada a la base de dades.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="update">Entitat modificada per actualitzar al context.</param>
        Task<TEntity> UpdateAsync(TKey key, TEntity update);

        #region Patch - Actualització parcial mitjançant Delta<TEntity> de "System.Web.Http.OData"

        ///// <summary>
        ///// Actualitza parcialment l'entitat modificada a la base de dades.
        ///// </summary>
        ///// <param name="key">Clau primària de l'entitat.</param>
        ///// <param name="patch">Canvis a l'entitat per actualitzar al context.</param>
        //TEntity Patch(TKey key, Delta<TEntity> patch);
        ///// <summary>
        ///// Actualitza parcialment l'entitat modificada a la base de dades.
        ///// </summary>
        ///// <remarks>Versió asíncrona.</remarks>
        ///// <param name="key">Clau primària de l'entitat.</param>
        ///// <param name="patch">Canvis a l'entitat per actualitzar al context.</param>
        //Task<TEntity> PatchAsync(TKey key, Delta<TEntity> patch);

        #endregion

        /// <summary>
        /// Esborra l'entitat de la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        void Delete(TKey key);
        /// <summary>
        /// Esborra l'entitat de la base de dades.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="key">Clau primària de l'entitat.</param>
        Task DeleteAsync(TKey key);
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        void CreateLink(TKey key, string navigationProperty, TKey relatedKey);
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        Task CreateLinkAsync(TKey key, string navigationProperty, TKey relatedKey);
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        void DeleteLink(TKey key, string navigationProperty, TKey relatedKey);
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        Task DeleteLinkAsync(TKey key, string navigationProperty, TKey relatedKey);
        ///// <summary>
        ///// Allibera recursos del context de la base de dades.
        ///// </summary>
        //void Dispose();
        /// <summary>
        /// Allibera recursos del context de la base de dades.
        /// </summary>
        /// <param name="disposing">Allibera o no recursos compartits.</param>
        void Dispose(bool disposing);
        //HttpResponseMessage Patch(TKey key, Delta<TEntity> patch);
        //TEntity PatchEntity(TKey key, System.Web.Http.OData.Delta<TEntity> patch);
        //HttpResponseMessage Post(TEntity entity);
        //HttpResponseMessage Put(TKey key, TEntity update);
        //HttpResponseMessage Get(TKey key);
        //HttpResponseMessage HandleUnmappedRequest(ODataPath odataPath);
        //void Initialize(HttpControllerContext controllerContext);
        //HttpResponseMessage Patch(TKey key, Delta<TEntity> patch);
    }
}
