﻿using System;
using System.Collections.Generic;
using System.Data.Entity;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Threading.Tasks;

namespace fgxavi.Common.Business.Repositories
{
    public class QueryableRepositoryImpl<TEntity> : IQueryableRepository<TEntity> where TEntity : class
    {
        protected internal DbContext dbContext;
        protected internal DbSet<TEntity> dbSet;
        public readonly IDictionary<string, PropertyInfo> Keys;
        public readonly IDictionary<string, PropertyInfo> Fields;
        protected QueryableRepositoryImpl()
        {
        }
        public QueryableRepositoryImpl(DbContext dbContext)
        {
            this.dbContext = dbContext;
            this.dbSet = dbContext.Set<TEntity>();
            Keys = dbContext.GetTableKeyColumns(typeof(TEntity));
            Fields = dbContext.GetTableColumns(typeof(TEntity));
        }
        public Task<int> CountAsync
        {
            get { return dbSet.CountAsync(); }
        }
        public Task<bool> ContainsAsync(TEntity item)
        {
            return dbSet.AnyAsync(t => t == item);
        }
        public Task<int> InsertAsync(TEntity item)
        {
            dbSet.Add(item);
            // Aplica els canvis del context a la base de dades de forma asíncrona.
            return dbContext.SaveChangesAsync();
        }
        public Task<int> UpdateAsync(TEntity item)
        {
            // Inclou l'entitat al context de la base de dades.
            var entity = dbSet.Attach(item);
            // Obté l'entrada de l'entitat del context de la base de dades.
            var entry = dbContext.Entry<TEntity>(entity);
            // Comprova l'estat de l'entrada de l'entitat al context.
            if (entry.State != EntityState.Unchanged)
                // L'entitat no existeix i no es pot modificar.
                return null;
            // Recorre per propietats de l'entitat excepte les claus primàries.
            foreach (var kv in Fields.Except(Keys))
            {
                // Obté el valor original i el valor nou de la propietat de l'entitat.
                var old_value = kv.Value.GetValue(entity);
                var new_value = kv.Value.GetValue(item);
                // Només modifica la propietat si el valor anterior i el nou són diferents.
                if (old_value != new_value)
                    // Estableix la propietat de l'entitat com a modificada perquè s'envii a la base de dades.
                    entry.Property(kv.Key).IsModified = true;
            }
            // Aplica els canvis del context a la base de dades de forma asíncrona.
            return dbContext.SaveChangesAsync();
        }
        public Task<int> DeleteAsync(TEntity item)
        {
            dbSet.Remove(item);
            // Aplica els canvis del context a la base de dades de forma asíncrona.
            return dbContext.SaveChangesAsync();
        }
        public IQueryable<TEntity> Find(string path = null)
        {
            return path == null ? dbSet : dbSet.Include(path);
        }
        public IQueryable<TEntity> Find(Func<TEntity, bool> predicate, string path = null)
        {
            return path == null ? dbSet : dbSet.Include(path)
                .Where(predicate).AsQueryable<TEntity>();
        }
        public IQueryable<TEntity> Find(Expression<Func<TEntity, bool>> predicate, string path = null)
        {
            return path == null ? dbSet : dbSet.Include(path)
                .Include(path).Where(predicate).AsQueryable<TEntity>();
        }
    }
}
