﻿using System;
using System.Collections.Generic;
using System.Data.Entity;
using System.Linq;
using System.Reflection;
using System.Threading.Tasks;

namespace fgxavi.Common.Business.Repositories
{
    public class RepositoryImpl<TEntity, TKey> : IRepository<TEntity, TKey>
        where TEntity : class
    {
        protected internal DbContext dbContext;
        protected internal DbSet<TEntity> dbSet;
        public Func<TEntity, TKey> KeySelector { get; set; }
        public IDictionary<string, PropertyInfo> Keys { get; private set; }
        public IDictionary<string, PropertyInfo> Fields { get; private set; }
        public IDictionary<string, PropertyInfo> Navigations { get; private set; }
        protected RepositoryImpl()
        {
        }
        public RepositoryImpl(DbContext dbContext)
        {
            // Apunta al context de dades i a les entitats del repositori.
            this.dbContext = dbContext;
            this.dbSet = dbContext.Set<TEntity>();
            // Indexa les claus primàries d'una entitat.
            Keys = dbContext.GetTableKeyColumns(typeof(TEntity));
            // Indexa els camps d'una entitat.
            Fields = dbContext.GetTableColumns(typeof(TEntity));
            // Indexa les propietats de navegació d'una entitat.
            Navigations = dbContext.GetTableNavigationColumns(typeof(TEntity));
            // Si l'entitat només conté una clau primària la retorna en la funció keySelector.
            if (Keys.Count == 1)
                KeySelector = new Func<TEntity, TKey>(
                    (entity) => (TKey)Keys.First().Value
                        .GetGetMethod().Invoke(entity, null));
        }
        /// <summary>
        /// Inicialitza la instància amb una funció que obté la clau primària d'una entitat.
        /// </summary>
        /// <remarks>Cal executar aquest mètode tan bon punt es crei la instància perquè el mètode GetKey retorni el valor correcte, en cas de no executar-se el mètode GetKey retornarà el valor per defecte del tipus de la clau primària.</remarks>
        public void Initialize(Func<TEntity, TKey> keySelector)
        {
            this.KeySelector = keySelector;
        }
        /// <summary>
        /// Obté l'entitat de tipus TEntity que té com a clau primària el tipus TKey facilitat.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <returns>Entitat coincident.</returns>
        public IQueryable<TEntity> GetAll()
        {
            // Retorna totes les entitats iguals del context.
            return dbSet;
        }
        /// <summary>
        /// Obté la clau primària de tipus TKey de l'entitat.
        /// </summary>
        /// <param name="entity">Entitat per a obtenir la clau primària.</param>
        /// <returns>Clau primària de l'entitat.</returns>
        public TEntity GetByKey(TKey key)
        {
            var entity = dbSet.Find(key);
            return entity;
        }
        /// <summary>
        /// Obté la clau primària de tipus TKey de l'entitat.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="entity">Entitat per a obtenir la clau primària.</param>
        /// <returns>Clau primària de l'entitat.</returns>
        public Task<TEntity> GetByKeyAsync(TKey key)
        {
            var entity = dbSet.FindAsync(key);
            return entity;
        }
        /// <summary>
        /// Obté la clau primària de tipus TKey de l'entitat.
        /// </summary>
        /// <remarks>Cal inicialitzar la instància de la classe amb el mètode Initialize per definir la funció que retorna la clau primària d'una entitat. En cas de no definir-se la funció es retorna el valor per defecte del tipus de la clau primària.</remarks>
        /// <param name="entity">Entitat per a obtenir la clau primària.</param>
        /// <returns>Clau primària de l'entitat.</returns>
        public TKey GetKey(TEntity entity)
        {
            if (KeySelector == null)
                return default(TKey);
            var key = KeySelector(entity);
            return key;
        }
        /// <summary>
        /// Obté una consulta IQueryable de totes les entitats de tipus TEntity de la base de dades.
        /// </summary>
        /// <remarks>No es realitza cap consulta a la base de dades fins que no s'executa la consulta IQueryable convertint-se en una enumeració IEnumerable.</remarks>
        /// <returns>Consulta IQueryable de totes les entitats de tipus TEntity.</returns>
        public TEntity Create(TEntity entity)
        {
            // Afegeix l'entitat al context de la base de dades.
            var entityAdded = dbSet.Add(entity);
            // Aplica els canvis del context a la base de dades.
            dbContext.SaveChanges();
            return entityAdded;
        }
        /// <summary>
        /// Obté una consulta IQueryable de totes les entitats de tipus TEntity de la base de dades.
        /// </summary>
        /// <remarks>
        /// No es realitza cap consulta a la base de dades fins que no s'executa la consulta IQueryable convertint-se en una enumeració IEnumerable.
        /// Versió asíncrona.
        /// </remarks>
        /// <returns>Consulta IQueryable de totes les entitats de tipus TEntity.</returns>
        public async Task<TEntity> CreateAsync(TEntity entity)
        {
            // Afegeix l'entitat al context de la base de dades.
            var entityAdded = dbSet.Add(entity);
            // Aplica els canvis del context a la base de dades.
            await dbContext.SaveChangesAsync();
            return entityAdded;
        }
        /// <summary>
        /// Actualitza l'entitat modificada a la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="update">Entitat modificada per actualitzar al context.</param>
        public TEntity Update(TKey key, TEntity update)
        {
            // Inclou l'entitat al context de la base de dades.
            var entity = dbSet.Attach(update);
            // Obté l'entrada de l'entitat del context de la base de dades.
            var entry = dbContext.Entry<TEntity>(entity);
            // Comprova l'estat de l'entrada de l'entitat al context.
            if (entry.State != EntityState.Unchanged)
                // L'entitat no existeix i no es pot modificar.
                return null;
            // Recorre per propietats de l'entitat excepte les claus primàries.
            foreach (var kv in Fields.Except(Keys))
            {
                // Obté el valor original i el valor nou de la propietat de l'entitat.
                var old_value = kv.Value.GetValue(entity);
                var new_value = kv.Value.GetValue(update);
                // Només modifica la propietat si el valor anterior i el nou són diferents.
                if (old_value != new_value)
                    // Estableix la propietat de l'entitat com a modificada perquè s'envii a la base de dades.
                    entry.Property(kv.Key).IsModified = true;
            }
            // Aplica els canvis del context a la base de dades.
            dbContext.SaveChanges();
            return entity;
        }
        /// <summary>
        /// Actualitza l'entitat modificada a la base de dades.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="update">Entitat modificada per actualitzar al context.</param>
        public async Task<TEntity> UpdateAsync(TKey key, TEntity update)
        {
            // Inclou l'entitat al context de la base de dades.
            var entity = dbSet.Attach(update);
            // Obté l'entrada de l'entitat del context de la base de dades.
            var entry = dbContext.Entry<TEntity>(entity);
            // Comprova l'estat de l'entrada de l'entitat al context.
            if (entry.State != EntityState.Unchanged)
                // L'entitat no existeix i no es pot modificar.
                return null;
            // Recorre per propietats de l'entitat excepte les claus primàries.
            foreach (var kv in Fields.Except(Keys))
            {
                // Obté el valor original i el valor nou de la propietat de l'entitat.
                var old_value = kv.Value.GetValue(entity);
                var new_value = kv.Value.GetValue(update);
                // Només modifica la propietat si el valor anterior i el nou són diferents.
                if (old_value != new_value)
                    // Estableix la propietat de l'entitat com a modificada perquè s'envii a la base de dades.
                    entry.Property(kv.Key).IsModified = true;
            }
            // Aplica els canvis del context a la base de dades.
            await dbContext.SaveChangesAsync();
            return entity;
        }

        #region Patch - Actualització parcial mitjançant Delta<TEntity> de "System.Web.Http.OData"

        ///// <summary>
        ///// Actualitza parcialment l'entitat modificada a la base de dades.
        ///// </summary>
        ///// <param name="key">Clau primària de l'entitat.</param>
        ///// <param name="patch">Canvis a l'entitat per actualitzar al context.</param>
        //public TEntity Patch(TKey key, Delta<TEntity> patch)
        //{
        //    // Obté l'entitat original de la base de dades.
        //    var entity = dbSet.Find(key);
        //    // Aplica els canvis a l'entitat del context.
        //    patch.Patch(entity);
        //    // Aplica els canvis del context a la base de dades.
        //    dbContext.SaveChanges();
        //    return entity;
        //}
        ///// <summary>
        ///// Actualitza parcialment l'entitat modificada a la base de dades.
        ///// </summary>
        ///// <remarks>Versió asíncrona.</remarks>
        ///// <param name="key">Clau primària de l'entitat.</param>
        ///// <param name="patch">Canvis a l'entitat per actualitzar al context.</param>
        //public async Task<TEntity> PatchAsync(TKey key, Delta<TEntity> patch)
        //{
        //    // Obté l'entitat original de la base de dades.
        //    var entity = await dbSet.FindAsync(key);
        //    // Aplica els canvis a l'entitat del context.
        //    patch.Patch(entity);
        //    // Aplica els canvis del context a la base de dades.
        //    await dbContext.SaveChangesAsync();
        //    return entity;
        //}

        #endregion

        /// <summary>
        /// Esborra l'entitat de la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        public void Delete(TKey key)
        {
            // Obté l'entitat de la base de dades.
            var entity = GetByKey(key);
            // Esborra l'entitat del context de la base de dades.
            dbSet.Remove(entity);
            // Aplica els canvis del context a la base de dades.
            dbContext.SaveChanges();
        }
        /// <summary>
        /// Esborra l'entitat de la base de dades.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="key">Clau primària de l'entitat.</param>
        public async Task DeleteAsync(TKey key)
        {
            // Obté l'entitat de la base de dades.
            var entity = await GetByKeyAsync(key);
            // Esborra l'entitat del context de la base de dades.
            dbSet.Remove(entity);
            // Aplica els canvis del context a la base de dades.
            await dbContext.SaveChangesAsync();
        }
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        public void CreateLink(TKey key, string navigationProperty, TKey relatedKey)
        {
            // Obté l'entitat principal mitjançant la clau primària.
            TEntity entity = dbSet.Find(key);
            // Obté l'entitat relacionada mitjançant la propietat de navegació i la clau primària relacionada.
            PropertyInfo propInfo = Navigations[navigationProperty];
            var relatedSet = dbContext.Set(propInfo.PropertyType);
            var relatedEntity = relatedSet.Find(relatedKey);
            // Estableix la relació entre ambdues entitats.
            if (typeof(ICollection<>).IsAssignableFrom(propInfo.PropertyType))
            {
                // Si la propietat de navegació és una col·lecció li afegeix l'entitat relacionada.
                var get = propInfo.GetGetMethod();
                if (!get.IsStatic && get.GetParameters().Length == 0) // skip indexed & static
                {
                    var collection = (ICollection<object>)get.Invoke(entity, null);
                    collection.Add(relatedEntity);
                }
            }
            else
            {
                // Si la propietat de navegació no és una col·lecció assigna l'entitat relacionada.
                propInfo.SetValue(entity, relatedEntity);
            }
            // Aplica els canvis del context a la base de dades.
            dbContext.SaveChanges();
        }
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        public async Task CreateLinkAsync(TKey key, string navigationProperty, TKey relatedKey)
        {
            // Obté l'entitat principal mitjançant la clau primària.
            TEntity entity = dbSet.Find(key);
            // Obté l'entitat relacionada mitjançant la propietat de navegació i la clau primària relacionada.
            PropertyInfo propInfo = Navigations[navigationProperty];
            var relatedSet = dbContext.Set(propInfo.PropertyType);
            var relatedEntity = await relatedSet.FindAsync(relatedKey);
            // Estableix la relació entre ambdues entitats.
            if (typeof(ICollection<>).IsAssignableFrom(propInfo.PropertyType))
            {
                // Si la propietat de navegació és una col·lecció li afegeix l'entitat relacionada.
                var get = propInfo.GetGetMethod();
                if (!get.IsStatic && get.GetParameters().Length == 0) // skip indexed & static
                {
                    var collection = (ICollection<object>)get.Invoke(entity, null);
                    collection.Add(relatedEntity);
                }
            }
            else
            {
                // Si la propietat de navegació no és una col·lecció assigna l'entitat relacionada.
                propInfo.SetValue(entity, relatedEntity);
            }
            // Aplica els canvis del context a la base de dades.
            await dbContext.SaveChangesAsync();
        }
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        public void DeleteLink(TKey key, string navigationProperty, TKey relatedKey)
        {
            // Obté l'entitat principal mitjançant la clau primària.
            TEntity entity = dbSet.Find(key);
            // Obté l'entitat relacionada mitjançant la propietat de navegació i la clau primària relacionada.
            PropertyInfo propInfo = Navigations[navigationProperty];
            var relatedSet = dbContext.Set(propInfo.PropertyType);
            var relatedEntity = relatedSet.Find(relatedKey);
            // Estableix la relació entre ambdues entitats.
            if (typeof(ICollection<>).IsAssignableFrom(propInfo.PropertyType))
            {
                // Si la propietat de navegació és una col·lecció n'elimina l'entitat relacionada.
                var get = propInfo.GetGetMethod();
                if (!get.IsStatic && get.GetParameters().Length == 0) // skip indexed & static
                {
                    var collection = (ICollection<object>)get.Invoke(entity, null);
                    collection.Remove(relatedEntity);
                }
            }
            else
            {
                // Si la propietat de navegació no és una col·lecció assigna el valor nul.
                propInfo.SetValue(entity, null);
            }
            // Aplica els canvis del context a la base de dades.
            dbContext.SaveChanges();
        }
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <remarks>Versió asíncrona.</remarks>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        public async Task DeleteLinkAsync(TKey key, string navigationProperty, TKey relatedKey)
        {
            // Obté l'entitat principal mitjançant la clau primària.
            TEntity entity = dbSet.Find(key);
            // Obté l'entitat relacionada mitjançant la propietat de navegació i la clau primària relacionada.
            PropertyInfo propInfo = Navigations[navigationProperty];
            var relatedSet = dbContext.Set(propInfo.PropertyType);
            var relatedEntity = await relatedSet.FindAsync(relatedKey);
            // Estableix la relació entre ambdues entitats.
            if (typeof(ICollection<>).IsAssignableFrom(propInfo.PropertyType))
            {
                // Si la propietat de navegació és una col·lecció n'elimina l'entitat relacionada.
                var get = propInfo.GetGetMethod();
                if (!get.IsStatic && get.GetParameters().Length == 0) // skip indexed & static
                {
                    var collection = (ICollection<object>)get.Invoke(entity, null);
                    collection.Remove(relatedEntity);
                }
            }
            else
            {
                // Si la propietat de navegació no és una col·lecció assigna el valor nul.
                propInfo.SetValue(entity, null);
            }
            // Aplica els canvis del context a la base de dades.
            await dbContext.SaveChangesAsync();
        }
        ///// <summary>
        ///// Allibera recursos del context de la base de dades.
        ///// </summary>
        //public void Dispose()
        //{
        //    this.dbContext.Dispose();
        //}
        /// <summary>
        /// Allibera recursos del context de la base de dades.
        /// </summary>
        /// <param name="disposing">Allibera o no recursos compartits.</param>
        public void Dispose(bool disposing)
        {
            this.dbContext.Dispose();
        }
    }
}
