﻿using System;

namespace fgxavi.Common.Helpers
{

    #region Interface IIdentityCard<T>

    /// <summary>
    /// Interfície genèrica d'una targeta d'identitat.
    /// </summary>
    /// <typeparam name="T">
    /// Tipus de dades que contindrà el nombre identificador de la targeta d'identitat.
    /// </typeparam>
    interface IIdentityCard<T>
    {

        /// <summary>
        /// Nombre identificador de la targeta d'identitat.
        /// </summary>
        T Number { get; set; }

        /// <summary>
        /// Propietat de lectura que retorna un nombre identificador
        /// aleatòri vàlid de la targeta d'identitat.
        /// </summary>
        T Random { get; }

        /// <summary>
        /// Funció lambda de lectura que valida un nombre identificador.
        /// </summary>
        Func<T, bool> Valid { get; }

    }

    #endregion

    #region Interface IIdentityCardWithChecksum<T>

    /// <summary>
    /// Interfície genèrica d'una targeta d'identitat amb codi de control.
    /// </summary>
    /// <typeparam name="T">
    /// Tipus de dades que contindrà el nombre identificador de la targeta d'identitat.
    /// </typeparam>
    /// <remarks>
    /// El tipus de dades del codi de control i del nombre identificador
    /// és el mateix que el el nombre de la targeta d'identitat.
    /// </remarks>
    interface IIdentityCardWithChecksum<T> : IIdentityCard<T>
    {

        /// <summary>
        /// Propietat de lectura amb el nombre identificador.
        /// </summary>
        T NumberCode { get; }

        /// <summary>
        /// Propietat de lectura amb el codi de control.
        /// </summary>
        T Checksum { get; }

        /// <summary>
        /// Funció lambda de lectura que calcula el codi de control.
        /// </summary>
        Func<T, T> ChecksumAlgorithm { get; }

        /// <summary>
        /// Funció lambda de lectura que extreu el nombre identificador
        /// d'una targeta d'identitat.
        /// </summary>
        Func<T, T> ExtractNumberCode { get; }

        /// <summary>
        /// Funció lambda de lectura que extreu el codi de control
        /// d'una targeta d'identitat.
        /// </summary>
        Func<T, T> ExtractChecksum { get; }

    }

    #endregion

    #region Class IdentityCardBase<T>

    /// <summary>
    /// Classe genèrica abstracta que implementa una targeta d'identitat.
    /// </summary>
    /// <typeparam name="T">
    /// Tipus de dades que contindrà el nombre identificador de la targeta d'identitat.
    /// </typeparam>
    public abstract class IdentityCardBase<T> : IIdentityCard<T>
    {

        /// <summary>
        /// Propietat estàtica protegida per utilitzar el mateix generador aleatòri 
        /// en totes les targetes d'identitat.
        /// </summary>
        protected static Random rnd = new Random();

        private T _number;

        /// <summary>
        /// Nombre identificador de la targeta d'identitat.
        /// </summary>
        public T Number
        {
            get { return _number; }
            set
            {
                if (!Valid(value))
                    throw new InvalidCastException();
                _number = value;
            }
        }

        /// <summary>
        /// Propietat de lectura que retorna un nombre identificador
        /// aleatòri vàlid de la targeta d'identitat.
        /// </summary>
        public abstract T Random { get; }

        /// <summary>
        /// Funció lambda de lectura que valida un nombre identificador.
        /// </summary>
        public abstract Func<T, bool> Valid { get; }

    }

    #endregion

    #region Class IdentityCardWithChecksumBase<T>

    /// <summary>
    /// Classe genèrica abstracta que implementa una targeta d'identitat amb codi de control.
    /// </summary>
    /// <typeparam name="T">
    /// Tipus de dades que contindrà el nombre identificador de la targeta d'identitat.
    /// </typeparam>
    /// <remarks>
    /// El tipus de dades del codi de control i del nombre identificador
    /// és el mateix que el el nombre de la targeta d'identitat.
    /// </remarks>
    public abstract class IdentityCardWithChecksumBase<T> : IdentityCardBase<T>, IIdentityCardWithChecksum<T>
    {

        /// <summary>
        /// Propietat de lectura amb el nombre identificador.
        /// </summary>
        public T NumberCode { get { return ExtractNumberCode(Number); } }

        /// <summary>
        /// Propietat de lectura amb el codi de control.
        /// </summary>
        public T Checksum { get { return ExtractChecksum(Number); } }

        /// <summary>
        /// Funció lambda abstracta de lectura que calcula el codi de control.
        /// </summary>
        public abstract Func<T, T> ChecksumAlgorithm { get; }

        /// <summary>
        /// Funció lambda abstracta de lectura que extreu el nombre identificador
        /// d'una targeta d'identitat.
        /// </summary>
        public abstract Func<T, T> ExtractNumberCode { get; }

        /// <summary>
        /// Funció lambda abstracta de lectura que extreu el codi de control
        /// d'una targeta d'identitat.
        /// </summary>
        public abstract Func<T, T> ExtractChecksum { get; }

    }

    #endregion

    #region Class DNI : IdentityCardBase<string>

    /// <summary>
    /// Implementació de la targeta d'identitat DNI.
    /// </summary>
    /// <remarks>
    /// El tipus de dades utilitzat per emmagatzemar el nombre identificador
    /// és una cadena de texte.
    /// </remarks>
    public class DNI : IdentityCardBase<string>
    {

        #region Order Random

        /// <summary>
        /// Propietat de lectura que retorna un nombre identificador
        /// aleatòri vàlid de la targeta d'identitat DNI.
        /// </summary>
        public override string Random
        {
            get
            {
                int num = rnd.Next(1000, 99999999);
                return num.ToString("00000000");
            }
        }

        #endregion

        #region Lambda property Valid

        /// <summary>
        /// Funció lambda de lectura que valida un nombre identificador de DNI.
        /// </summary>
        public override Func<string, bool> Valid
        {
            get
            {
                return value =>
                {
                    // Comprova que el nombre del DNI sigui correcte:
                    int num;
                    return (value != null)
                        && (value.Length == 8)
                        && (!int.TryParse(value, out num));
                };
            }
        }

        #endregion

    }

    #endregion

    #region Class NIF : IdentityCardWithChecksumBase<string>

    /// <summary>
    /// Implementació de la targeta d'identitat amb codi de control NIF.
    /// </summary>
    /// <remarks>
    /// El tipus de dades utilitzat per emmagatzemar el nombre identificador
    /// és una cadena de texte. S'utilitza la classe DNI per validar i generar el DNI.
    /// </remarks>
    public class NIF : IdentityCardWithChecksumBase<string>
    {

        #region Constants

        /// <summary>
        /// Frase d'encriptació per correspondència utilitzada pel codi de control del NIF.
        /// </summary>
        public const string EncryptionPassphrase = "TRWAGMYFPDXBNJZSQVHLCKE";

        #endregion

        #region Order Random

        /// <summary>
        /// Propietat de lectura que retorna un nombre identificador
        /// aleatòri vàlid de la targeta d'identitat amb codi de control NIF.
        /// </summary>
        public override string Random
        {
            get
            {
                string numberCode = new DNI().Random;
                return numberCode + ChecksumAlgorithm(numberCode);
            }
        }

        #endregion

        #region Lambda property Valid

        /// <summary>
        /// Funció lambda de lectura que valida un nombre identificador de NIF.
        /// </summary>
        public override Func<string, bool> Valid
        {
            get
            {
                return value =>
                {
                    try
                    {
                        // Comprova que el nombre del DNI sigui correcte:
                        string numberCode = ExtractNumberCode(value);
                        if (!new DNI().Valid(numberCode))
                            return false;
                        // Comprova que la lletra del NIF sigui correcte:
                        string checksum = ExtractChecksum(value);
                        return checksum == ChecksumAlgorithm(numberCode);
                    }
                    catch (Exception)
                    {
                        return false;
                    }
                };
            }
        }

        #endregion

        #region Lambda property ChecksumAlgorithm

        /// <summary>
        /// Funció lambda de lectura que calcula el codi de control del NIF.
        /// </summary>
        public override Func<string, string> ChecksumAlgorithm
        {
            get
            {
                return value =>
                {
                    int numCode = int.Parse(value) % 23;
                    return EncryptionPassphrase.Substring(numCode, 1);
                };
            }
        }

        #endregion

        #region Lambda property ExtractNumberCode

        /// <summary>
        /// Funció lambda de lectura que extreu el nombre identificador
        /// d'una targeta d'identitat.
        /// </summary>
        public override Func<string, string> ExtractNumberCode
        {
            get
            {
                return value =>
                {
                    return value.Substring(0, value.Length - 1);
                };
            }
        }

        #endregion

        #region Lambda property ExtractChecksum

        /// <summary>
        /// Funció lambda abstracta de lectura que extreu el codi de control
        /// d'una targeta d'identitat.
        /// </summary>
        public override Func<string, string> ExtractChecksum
        {
            get
            {
                return value =>
                {
                    return value.Substring(value.Length - 1);
                };
            }
        }

        #endregion

    }

    #endregion

    #region Class NIE : IdentityCardWithChecksumBase<string>

    /// <summary>
    /// Implementació de la targeta d'identitat amb codi de control NIE.
    /// </summary>
    /// <remarks>
    /// El tipus de dades utilitzat per emmagatzemar el nombre identificador
    /// és una cadena de texte.
    /// </remarks>
    public class NIE : IdentityCardWithChecksumBase<string>
    {

        #region Constants

        /// <summary>
        /// Frase d'encriptació per correspondència utilitzada pel codi de control del NIE.
        /// </summary>
        public const string EncryptionPassphrase = "TRWAGMYFPDXBNJZSQVHLCKE";

        /// <summary>
        /// Caràcters vàlids utilitzats a principi del nombre identificador del NIE.
        /// </summary>
        public const string ValidStartCharacters = "XYZ";

        #endregion

        #region Order Random

        /// <summary>
        /// Propietat de lectura que retorna un nombre identificador
        /// aleatòri vàlid de la targeta d'identitat amb codi de control NIE.
        /// </summary>
        public override string Random
        {
            get
            {
                // Obté el primer caràcter vàlid per un NIE:
                string startCharacter = ValidStartCharacters.Substring(rnd.Next(ValidStartCharacters.Length), 1);
                // Obté un nombre vàlid per un NIE:
                int num = rnd.Next(1000, 9999999);
                // Nombre identificador vàlid per un NIE:
                string numberCode = startCharacter + num.ToString("0000000");
                // retorna el nombre identificador vàlid del NIE amb el seu codi de control
                return numberCode + ChecksumAlgorithm(numberCode);
            }
        }

        #endregion

        #region Lambda property Valid

        /// <summary>
        /// Funció lambda de lectura que valida un nombre identificador de NIE.
        /// </summary>
        public override Func<string, bool> Valid
        {
            get
            {
                return value =>
                {
                    try
                    {
                        // Comprova que el nombre del NIE sigui correcte:
                        string numberCode = ExtractNumberCode(value);
                        int num;
                        if (value == null
                            || ValidStartCharacters.Contains(numberCode.Substring(0, 1))
                            || value.Length != 8
                            || int.TryParse(numberCode.Substring(1, 7), out num))
                            return false;
                        // Comprova que la lletra del NIE sigui correcte:
                        string checksum = ExtractChecksum(value);
                        return checksum == ChecksumAlgorithm(numberCode);
                    }
                    catch (Exception)
                    {
                        return false;
                    }
                };
            }
        }

        #endregion

        #region Lambda property ChecksumAlgorithm

        /// <summary>
        /// Funció lambda de lectura que calcula el codi de control del NIE.
        /// </summary>
        public override Func<string, string> ChecksumAlgorithm
        {
            get
            {
                return value =>
                {
                    if (!ValidStartCharacters.Contains(value.Substring(0, 1)))
                        throw new InvalidCastException();
                    int numCode = (ValidStartCharacters.IndexOf(value.Substring(0, 1)) * 10000000
                        + int.Parse(value.Substring(1, 7))) % 23;
                    return EncryptionPassphrase.Substring(numCode, 1);
                };
            }
        }

        #endregion

        #region Lambda property ExtractNumberCode

        /// <summary>
        /// Funció lambda de lectura que extreu el nombre identificador
        /// d'una targeta d'identitat.
        /// </summary>
        public override Func<string, string> ExtractNumberCode
        {
            get
            {
                return value =>
                {
                    return value.Substring(0, value.Length - 1);
                };
            }
        }

        #endregion

        #region Lambda property ExtractChecksum

        /// <summary>
        /// Funció lambda abstracta de lectura que extreu el codi de control
        /// d'una targeta d'identitat.
        /// </summary>
        public override Func<string, string> ExtractChecksum
        {
            get
            {
                return value =>
                {
                    return value.Substring(value.Length - 1);
                };
            }
        }

        #endregion

    }

    #endregion

}
