﻿using fgxavi.Common.Business.Repositories;
using fgxavi.Common.Helpers;
using Microsoft.Practices.Unity;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;
using System.Web.Http.Controllers;
using System.Web.Http.OData;
using System.Web.Http.OData.Query;

namespace fgxavi.Common.Services
{
    public abstract class AsyncRepositoryBaseController<TEntity, TKey> : AsyncEntitySetController<TEntity, TKey>
        where TEntity : class
    {
        protected IRepository<TEntity, TKey> repository;
        //public AsyncRepositoryBaseController()
        //{
        //    // Obté la instància del contenidor UnityContainer actual.
        //    var container = Singleton<IUnityContainer, UnityContainer>.Instance;
        //    // Obté el repositori d'aquesta entitat que comunicarà amb la base de dades.
        //    repository = container.Resolve<IRepository<TEntity, TKey>>();
        //}
        public AsyncRepositoryBaseController(IRepository<TEntity, TKey> repository)
        {
            this.repository = repository;
        }
        ///// <summary>
        ///// Implementació necessària de la classe base per inicialitzar correctament el repositori.
        ///// </summary>
        ///// <remarks>El repositori necessita necessita definir la funció que retorna la clau primària del tipus de l'entitat. Per fer-ho s'ha d'executar this.repository.Initialize(keySelector).</remarks>
        //protected abstract override void Initialize(HttpControllerContext controllerContext);
        /// <summary>
        /// Obté una consulta IQueryable de totes les entitats de tipus TEntity de la base de dades.
        /// </summary>
        /// <remarks>No es realitza cap consulta a la base de dades fins que no s'executa la consulta IQueryable convertint-se en una enumeració IEnumerable.</remarks>
        /// <returns>Consulta IQueryable de totes les entitats de tipus TEntity.</returns>
        //[ControlledQueryable(PageSize=10)]
        public IEnumerable<TEntity> GetAsync(ODataQueryOptions<TEntity> queryOptions)
        {
            // Obté la consulta del repositori.
            var query = repository.GetAll();
            // Aplica les opcions de consulta sobre el total d'entitats.
            query = queryOptions.ApplyTo(query) as IQueryable<TEntity>;
            // Retorna l'enumeració de resultats obtinguts.
            var result = query.AsEnumerable();
            return result;
        }
        //public override Task<IEnumerable<TEntity>> Get()
        //{
        //    var result = repository.GetAll();
        //    return Task.Run<IEnumerable<TEntity>>(() => result.ToList());
        //}
        /// <summary>
        /// Obté l'entitat de tipus TEntity que té com a clau primària el tipus TKey facilitat.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <returns>Entitat coincident.</returns>
        protected override Task<TEntity> GetEntityByKeyAsync(TKey key)
        {
            var result = repository.GetByKeyAsync(key);
            return result;
        }
        /// <summary>
        /// Obté la clau primària de tipus TKey de l'entitat.
        /// </summary>
        /// <param name="entity">Entitat per a obtenir la clau primària.</param>
        /// <returns>Clau primària de l'entitat.</returns>
        protected override TKey GetKey(TEntity entity)
        {
            var key = repository.GetKey(entity);
            return key;
        }
        /// <summary>
        /// Crea una nova entitat a la base de dades.
        /// </summary>
        /// <param name="entity">Entitat per afegir al context.</param>
        /// <returns>Entitat creada.</returns>
        protected override Task<TEntity> CreateEntityAsync(TEntity entity)
        {
            var result = repository.CreateAsync(entity);
            return result;
        }
        /// <summary>
        /// Actualitza l'entitat modificada a la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="update">Entitat modificada per actualitzar al context.</param>
        protected override Task<TEntity> UpdateEntityAsync(TKey key, TEntity update)
        {
            var entity = repository.UpdateAsync(key, update);
            return entity;
        }
        /// <summary>
        /// Actualitza parcialment l'entitat modificada a la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="patch">Canvis a l'entitat per actualitzar al context.</param>
        protected async override Task<TEntity> PatchEntityAsync(TKey key, Delta<TEntity> patch)
        {
            // Obté l'entitat original de la base de dades.
            var entity = await repository.GetByKeyAsync(key);
            // Aplica els canvis a l'entitat del context.
            patch.Patch(entity);
            // Actualitza l'entitat a la base de dades.
            entity = await repository.UpdateAsync(key, entity);
            //var entity = repository.PatchAsync(key, patch);
            return entity;
        }
        /// <summary>
        /// Esborra l'entitat de la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        public override Task Delete(TKey key)
        {
            return repository.DeleteAsync(key);
        }
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="link">Enllaç a l'entitat relacionada.</param>
        public override Task CreateLink(TKey key, string navigationProperty, Uri link)
        {
            TKey relatedKey = Request.GetKeyFromLinkUri<TKey>(link);
            return repository.CreateLinkAsync(key, navigationProperty, relatedKey);
        }
        /// <summary>
        /// Esborra un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        public override Task DeleteLink(TKey key, string relatedKey, string navigationProperty)
        {
            TKey related = relatedKey.ChangeType<TKey>();
            return repository.DeleteLinkAsync(key, navigationProperty, related);
        }
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="link">Enllaç a l'entitat relacionada.</param>
        public override Task DeleteLink(TKey key, string navigationProperty, Uri link)
        {
            TKey relatedKey = Request.GetKeyFromLinkUri<TKey>(link);
            return repository.DeleteLinkAsync(key, navigationProperty, relatedKey);
        }
        /// <summary>
        /// Allibera recursos del context de la base de dades.
        /// </summary>
        /// <param name="disposing">Allibera o no recursos compartits.</param>
        protected override void Dispose(bool disposing)
        {
            repository.Dispose(disposing);
            base.Dispose(disposing);
        }
        /// <summary>
        /// Retorna el nombre de resultats obtinguts amb la consulta OData.
        /// </summary>
        /// <param name="queryOptions">Opcions de consulta.</param>
        /// <returns>Nombre de resultats.</returns>
        public Task<int> GetCountAsync(ODataQueryOptions<TEntity> queryOptions)
        {
            // Obté la consulta del repositori.
            var query = repository.GetAll();
            // Aplica les opcions de consulta sobre el total d'entitats.
            query = queryOptions.ApplyTo(query) as IQueryable<TEntity>;
            // Obté el nombre de coincidències de la consulta de forma asíncrona.
            Task<int> count = Task<int>.Run(() => query.Count());
            // Retorna el nombre de resultats de la consulta.
            return count;
        }
    }
}
