﻿using fgxavi.Common.Business.Repositories;
using fgxavi.Common.Helpers;
using System;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Web.Http.OData;
using System.Web.Http.OData.Query;

namespace fgxavi.Common.Services
{
    public abstract class RepositoryBaseController<TEntity, TKey> : EntitySetController<TEntity, TKey>
        where TEntity : class
    {
        protected IRepository<TEntity, TKey> repository;
        //public RepositoryBaseController()
        //{
        //    // Obté la instància del contenidor UnityContainer actual.
        //    var container = Singleton<IUnityContainer, UnityContainer>.Instance;
        //    // Obté el repositori d'aquesta entitat que comunicarà amb la base de dades.
        //    repository = container.Resolve<IRepository<TEntity, TKey>>();
        //}
        public RepositoryBaseController(IRepository<TEntity, TKey> repository)
        {
            this.repository = repository;
        }
        ///// <summary>
        ///// Implementació de la classe base per inicialitzar correctament el repositori.
        ///// </summary>
        ///// <remarks>El repositori necessita necessita definir la funció que retorna la clau primària del tipus de l'entitat. Per fer-ho s'ha d'executar this.repository.Initialize(keySelector).</remarks>
        //protected abstract override void Initialize(HttpControllerContext controllerContext);
        /// <summary>
        /// Obté una consulta IQueryable de totes les entitats de tipus TEntity de la base de dades.
        /// </summary>
        /// <remarks>No es realitza cap consulta a la base de dades fins que no s'executa la consulta IQueryable convertint-se en una enumeració IEnumerable.</remarks>
        /// <returns>Consulta IQueryable de totes les entitats de tipus TEntity.</returns>
        public override IQueryable<TEntity> Get()
        {
            var result = repository.GetAll();
            return result;
        }
        /// <summary>
        /// Obté l'entitat de tipus TEntity que té com a clau primària el tipus TKey facilitat.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <returns>Entitat coincident.</returns>
        protected override TEntity GetEntityByKey(TKey key)
        {
            var result = repository.GetByKey(key);
            return result;
        }
        /// <summary>
        /// Obté la clau primària de tipus TKey de l'entitat.
        /// </summary>
        /// <param name="entity">Entitat per a obtenir la clau primària.</param>
        /// <returns>Clau primària de l'entitat.</returns>
        protected override TKey GetKey(TEntity entity)
        {
            var key = repository.GetKey(entity);
            return key;
        }
        /// <summary>
        /// Crea una nova entitat a la base de dades.
        /// </summary>
        /// <param name="entity">Entitat per afegir al context.</param>
        /// <returns>Entitat creada.</returns>
        protected override TEntity CreateEntity(TEntity entity)
        {
            var result = repository.Create(entity);
            return result;
        }
        /// <summary>
        /// Actualitza parcialment l'entitat modificada a la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="patch">Canvis a l'entitat per actualitzar al context.</param>
        protected override TEntity PatchEntity(TKey key, Delta<TEntity> patch)
        {
            // Obté l'entitat original de la base de dades.
            var entity = repository.GetByKey(key);
            // Aplica els canvis a l'entitat del context.
            patch.Patch(entity);
            // Actualitza l'entitat a la base de dades.
            entity = repository.Update(key, entity);
            //var entity = repository.Patch(key, patch);
            return entity;
        }
        /// <summary>
        /// Actualitza l'entitat modificada a la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="update">Entitat modificada per actualitzar al context.</param>
        protected override TEntity UpdateEntity(TKey key, TEntity update)
        {
            var entity = repository.Update(key, update);
            return entity;
        }
        /// <summary>
        /// Esborra l'entitat de la base de dades.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        public override void Delete(TKey key)
        {
            repository.Delete(key);
        }
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="link">Enllaç a l'entitat relacionada.</param>
        public override void CreateLink(TKey key, string navigationProperty, Uri link)
        {
            TKey relatedKey = Request.GetKeyFromLinkUri<TKey>(link);
            repository.CreateLink(key, navigationProperty, relatedKey);
        }
        /// <summary>
        /// Esborra un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="relatedKey">Clau primària de l'entitat relacionada.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        public override void DeleteLink(TKey key, string relatedKey, string navigationProperty)
        {
            TKey related = relatedKey.ChangeType<TKey>();
            repository.DeleteLink(key, navigationProperty, related);
        }
        /// <summary>
        /// Afegeix un lligam entre dues entitats relacionades per propietats de navegació.
        /// </summary>
        /// <param name="key">Clau primària de l'entitat.</param>
        /// <param name="navigationProperty">Propietat de navegació.</param>
        /// <param name="link">Enllaç a l'entitat relacionada.</param>
        public override void DeleteLink(TKey key, string navigationProperty, Uri link)
        {
            TKey relatedKey = Request.GetKeyFromLinkUri<TKey>(link);
            repository.DeleteLink(key, navigationProperty, relatedKey);
        }
        /// <summary>
        /// Allibera recursos del context de la base de dades.
        /// </summary>
        /// <param name="disposing">Allibera o no recursos compartits.</param>
        protected override void Dispose(bool disposing)
        {
            repository.Dispose(disposing);
            base.Dispose(disposing);
        }
        /// <summary>
        /// Retorna el nombre de resultats obtinguts amb la consulta OData.
        /// </summary>
        /// <param name="queryOptions">Opcions de consulta.</param>
        /// <returns>Nombre de resultats.</returns>
        public int GetCount(ODataQueryOptions<TEntity> queryOptions)
        {
            // Aplica les opcions de consulta abans d'obtenir el nombre de resultats.
            IQueryable<TEntity> queryResults = queryOptions.ApplyTo(Get()) as IQueryable<TEntity>;
            // Obté el nombre de coincidències de la consulta.
            int count = queryResults.Count();
            return count;
        }

        #region Exemple d'implementació CRUD en propietats de navegació:

        ///// <summary>
        ///// Support for /CurrentEntities(1)/RelatedEntities
        ///// </summary>
        //public IQueryable<TEntityRelated> GetRelatedEntities([FromODataUri] TKey key)
        //{
        //    var query = repository.GetAll().Where(p => p.KeyProperty == key).SelectMany(p => p.NavigationProperty);
        //    return query;
        //}

        ///// <summary>
        ///// Support for POST /CurrentEntities(1)/RelatedEntities
        ///// </summary>
        //public HttpResponseMessage PostProducts([FromODataUri] TKey key, TEntityRelated relatedEntity)
        //{
        //    var relatedKey = relatedEntity.KeyProperty;
        //    this.CreateLink(key, "navigationProperty", relatedKey);
        //    var odatalink = Url.ODataLink(
        //        new EntitySetPathSegment("RelatedEntities"),
        //        new KeyValuePathSegment(relatedKey.ToString()));
        //    var relatedEntityUri = new Uri(odatalink);
        //    var response = Request.CreateResponse(
        //        HttpStatusCode.Created,
        //        relatedEntity);
        //    response.Headers.Location = relatedEntityUri;
        //    return response;
        //}

        #endregion

    }
}
