﻿using System;
using System.Linq;
using System.Collections.Generic;

namespace fgxavi.ListView
{
    public enum ListViewFactories
    {
        ArrayFactory,
        JsonFactory,
        LinqFactory,
        ODataFactory
    }
    public interface IRefresh
    {
        /// <summary>
        /// Actualitza les propietats del objecte.
        /// </summary>
        void Refresh();
    }
    public interface IQueryOptions<TQuery, TKey, TValue>
    {

        #region All
        TQuery Query { get; }
        Func<IQueryOptions<TQuery, TKey, TValue>, TQuery> QueryBuilder { get; }
        #endregion

        #region Paged
        TQuery PagedQuery { get; }
        IDictionary<TKey, TValue> PagedParameters { get; }
        Func<IPagedQueryOption, IDictionary<TKey, TValue>> PagedParametersBuilder { get; }
        #endregion

        #region Sorted
        TQuery SortedQuery { get; }
        IDictionary<TKey, TValue> SortedParameters { get; set; }
        Func<ISortedQueryOption, IDictionary<TKey, TValue>> SortedParametersBuilder { get; }
        #endregion

    }
    public class QueryOptions<TQuery, TKey, TValue> : IQueryOptions<TQuery, TKey, TValue>
    {

        #region All
        public TQuery Query { get; set; }
        public Func<IQueryOptions<TQuery, TKey, TValue>, TQuery> QueryBuilder { get; set; }
        #endregion

        #region Paged
        public TQuery PagedQuery { get; set; }
        public IDictionary<TKey, TValue> PagedParameters { get; set; }
        public Func<IPagedQueryOption, IDictionary<TKey, TValue>> PagedParametersBuilder { get; set; }
        #endregion

        #region Sorted
        public TQuery SortedQuery { get; set; }
        public IDictionary<TKey, TValue> SortedParameters { get; set; }
        public Func<ISortedQueryOption, IDictionary<TKey, TValue>> SortedParametersBuilder { get; set; }
        #endregion
    }
    public interface IListViewFactory<TSource, TQuery, TQueryKey, TQueryValue, TConnection, TElement>
    {
        /// <summary>
        /// Paràmetres de consulta d'una font de dades.
        /// </summary>
        IQueryOptions<TQuery, TQueryKey, TQueryValue> QueryOptions { get; }
        /// <summary>
        /// Funció que construeix la connexió amb la font de dades.
        /// </summary>
        Func<TSource, IQueryOptions<TQuery, TQueryKey, TQueryValue>, TConnection> ConnectionBuilder { get; }
        /// <summary>
        /// Acció que actualitza un ViewModel a partir d'una connexió amb la font de dades.
        /// </summary>
        Action<TConnection, IListViewModel<TElement>> Connector { get; }
    }
    public abstract class ListViewBaseFactory<TSource, TQuery, TQueryKey, TQueryValue, TConnection, TElement>
        : IListViewFactory<TSource, TQuery, TQueryKey, TQueryValue, TConnection, TElement>
    {
        public const int MaxTop = 100;
        public const int DefaultTop = 10;
        public ListViewBaseFactory()
        {
            this.QueryOptions = this.buildQueryOptions();
            this.ConnectionBuilder = (source, queryoptions) => this.buildConnection(source, queryoptions);
            this.Connector = (connection, vm) => this.binding(connection, vm);
        }
        protected abstract QueryOptions<TQuery, TQueryKey, TQueryValue> buildQueryOptions();
        protected abstract TConnection buildConnection(TSource source, IQueryOptions<TQuery, TQueryKey, TQueryValue> queryoptions);
        protected abstract void binding(TConnection conn, IListViewModel<TElement> vm);
        /// <summary>
        /// Paràmetres de consulta d'una font de dades.
        /// </summary>
        public IQueryOptions<TQuery, TQueryKey, TQueryValue> QueryOptions { get; private set; }
        /// <summary>
        /// Funció que construeix la connexió amb la font de dades.
        /// </summary>
        public Func<TSource, IQueryOptions<TQuery, TQueryKey, TQueryValue>, TConnection> ConnectionBuilder { get; private set; }
        /// <summary>
        /// Acció que actualitza un ViewModel a partir d'una connexió amb la font de dades.
        /// </summary>
        public Action<TConnection, IListViewModel<TElement>> Connector { get; private set; }
    }
    public class ListViewODataFactory<TElement>
        : ListViewBaseFactory<string, string, string, string, string, TElement>
    {
        protected override QueryOptions<string, string, string> buildQueryOptions()
        {
            var queryoptions = new QueryOptions<string, string, string>()
            {
                QueryBuilder = (qo) =>
                {
                    var parameters = new Dictionary<string, string>()
                        .Union(qo.PagedParameters)
                        .Union(qo.SortedParameters);
                    var subqueries = parameters.Select(kv => string.Format("${0}={1}", kv.Key, kv.Value));
                    var query = string.Join("&", subqueries);
                    return query;
                },
                PagedParametersBuilder = (vm) =>
                {
                    var parameters = new Dictionary<string, string>();
                    parameters.Add("inlinecount", "allpages");
                    var topValid = vm.Top < 1 || vm.Top > MaxTop
                        ? DefaultTop : vm.Top;
                    parameters.Add("top", topValid.ToString());
                    var skipValid = vm.Skip < 0 ? 0 : vm.Skip;
                    if (skipValid > 0)
                        parameters.Add("skip", skipValid.ToString());
                    return parameters;
                },
                SortedParametersBuilder = (vm) =>
                {
                    var parameters = new Dictionary<string, string>();
                    return parameters;
                }
            };
            return queryoptions;
        }
        protected override string buildConnection(string source, IQueryOptions<string, string, string> queryoptions)
        {
            var connection = string.Format("{0}?{1}", source, queryoptions.Query);
            return connection;
        }
        protected override void binding(string conn, IListViewModel<TElement> vm)
        {
            // TODO falta implementar la connexió amb el servidor OData.
            // La següent és la implementació en les classes javascript:
            //Connector: function (connection) {
            //    var self = this;
            //    $.getJSON(connection, function (data) {
            //        self.vm().TotalItemCount($(data).attr('odata.count'));
            //        self.vm().Items(data.value);
            //    });
            //}
        }
    }
}
