﻿using System;
using System.Collections.Generic;

namespace fgxavi.ListView
{
    public class ListViewHelper
    {
        /// <summary>
        /// Obté el nombre total de pàgines.
        /// </summary>
        /// <param name="pageSize">Mida de les pàgines.</param>
        /// <param name="totalItemCount">Nombre total d'elements.</param>
        /// <returns>Nombre de pàgines.</returns>
        public static int PageCount(int pageSize, int totalItemCount)
        {
            return (int)Math.Ceiling((double)totalItemCount / (double)pageSize);
        }
        /// <summary>
        /// Corregeix un índex d'element incorrecte.
        /// </summary>
        /// <param name="item">Índex d'element sense verificar.</param>
        /// <param name="totalItemCount">Nombre total d'elements.</param>
        /// <returns>Índex d'element vàlid.</returns>
        public static int NormalizeItem(int item, int totalItemCount)
        {
            return Math.Max(Math.Min(item, totalItemCount), 1);
        }
        /// <summary>
        /// Corregeix un índex de pàgina incorrecte.
        /// </summary>
        /// <param name="pageNumber">Índex de pàgina sense verificar.</param>
        /// <param name="pageCount">Nombre total de pàgines.</param>
        /// <returns>Índex de pàgina vàlid.</returns>
        public static int NormalizePage(int pageNumber, int pageCount)
        {
            return Math.Max(Math.Min(pageNumber, pageCount), 1);
        }
        /// <summary>
        /// Corregeix un índex de pàgina incorrecte.
        /// </summary>
        /// <param name="item">Índex de pàgina sense verificar.</param>
        /// <param name="pageSize">Mida de les pàgines.</param>
        /// <param name="totalItemCount">Nombre total d'elements.</param>
        /// <returns>Índex de pàgina vàlid.</returns>
        public static int NormalizePage(int pageNumber, int pageSize, int totalItemCount)
        {
            var pageCount = PageCount(pageSize, totalItemCount);
            return NormalizePage(pageNumber, pageCount);
        }
        /// <summary>
        /// Calcula l'índex del primer element de la pàgina actual.
        /// </summary>
        /// <param name="pageNumber">Pàgina actual.</param>
        /// <param name="pageSize">Mida de les pàgines.</param>
        /// <param name="totalItemCount">Nombre total d'elements.</param>
        /// <returns>Índex del primer element de la pàgina actual.</returns>
        public static int? FirstItemOnPage(int pageNumber, int pageSize, int totalItemCount)
        {
            return totalItemCount == 0 ? null : (int?)Math.Min(totalItemCount, (pageNumber - 1) * pageSize + 1);
        }
        /// <summary>
        /// Calcula l'índex de l'últim element de la pàgina actual.
        /// </summary>
        /// <param name="pageNumber">Pàgina actual.</param>
        /// <param name="pageSize">Mida de les pàgines.</param>
        /// <param name="totalItemCount">Nombre total d'elements.</param>
        /// <returns>Índex de l'últim element de la pàgina actual.</returns>
        public static int? LastItemOnPage(int pageNumber, int pageSize, int totalItemCount)
        {
            return totalItemCount == 0 ? null : (int?)Math.Max(totalItemCount, pageNumber * pageSize - 1);
        }
        /// <summary>
        /// Obté els índex dels elements de la pàgina actual.
        /// </summary>
        /// <param name="pageNumber">Pàgina actual.</param>
        /// <param name="pageSize">Mida de les pàgines.</param>
        /// <param name="totalItemCount">Nombre total d'elements.</param>
        /// <returns>Índex del primer element de la pàgina actual.</returns>
        public static int[] ItemsOnPage(int pageNumber, int pageSize, int totalItemCount)
        {
            var min = FirstItemOnPage(pageNumber, pageSize, totalItemCount);
            var max = LastItemOnPage(pageNumber, pageSize, totalItemCount);
            var items = new List<int>();
            if (min.HasValue && max.HasValue)
                for (int i = min.Value; i <= max.Value; i++)
                    items.Add(i);
            return items.ToArray();
        }
        /// <summary>
        /// Obté un array d'enters amb l'interval de pàgines del navegador.
        /// </summary>
        /// <param name="pageNumber">Pàgina actual.</param>
        /// <param name="pageCount">Nombre total de pàgines.</param>
        /// <param name="maxRange">Nombre màxim de pàgines del rang.</param>
        /// <returns>Array d'enters amb l'interval de pàgines del navegador.</returns>
        public static int[] RangeNavigator(int pageNumber, int pageCount, int maxRange)
        {
            var interval = new List<int>();
            var min = (int)Math.Max(Math.Floor((decimal)pageNumber / (decimal)maxRange) * maxRange - 1, 1);
            var max = (int)Math.Min((Math.Floor((decimal)pageNumber / (decimal)maxRange) + 1) * maxRange, pageCount);
            for (int i = min; i <= max; i++)
                interval.Add(i);
            return interval.ToArray();
        }
        /// <summary>
        /// Obté un array d'enters amb l'interval de pàgines del navegador.
        /// </summary>
        /// <param name="pageNumber">Pàgina actual.</param>
        /// <param name="pageSize">Mida de les pàgines.</param>
        /// <param name="totalItemCount">Nombre total d'elements.</param>
        /// <param name="maxRange">Nombre màxim de pàgines del rang.</param>
        /// <returns>Array d'enters amb l'interval de pàgines del navegador.</returns>
        public static int[] RangeNavigator(int pageNumber, int pageSize, int totalItemCount, int maxRange)
        {
            var pageCount = PageCount(pageSize, totalItemCount);
            return RangeNavigator(pageNumber, pageCount, maxRange);
        }
        /// <summary>
        /// Obté la pàgina de navegació per l'esquerra a un altre rang de pàgines.
        /// </summary>
        /// <param name="rangeNavigator">Array d'enters amb l'interval de pàgines del navegador.</param>
        /// <returns>Pàgina de navegació a l'esquerra o nul si no n'hi ha.</returns>
        public static int? LeftNavigator(int[] rangeNavigator)
        {
            if (rangeNavigator == null || rangeNavigator.Length == 0)
                return null;
            var left = rangeNavigator[0] - 1;
            return left >= 1 ? (int?)left : null;
        }
        /// <summary>
        /// Obté la pàgina de navegació per la dreta a un altre rang de pàgines.
        /// </summary>
        /// <param name="rangeNavigator">Array d'enters amb l'interval de pàgines del navegador.</param>
        /// <param name="pageCount">Nombre total de pàgines.</param>
        /// <returns>Pàgina de navegació a la dreta o nul si no n'hi ha.</returns>
        public static int? RightNavigator(int[] rangeNavigator, int pageCount)
        {
            if (rangeNavigator == null || rangeNavigator.Length == 0)
                return null;
            var right = rangeNavigator[rangeNavigator.Length - 1] + 1;
            return right <= pageCount ? (int?)right : null;
        }
    }
}
