﻿using System.Collections.Generic;
using System.ComponentModel;

namespace fgxavi.ListView
{
    public interface IField
    {
        /// <summary>
        /// Nom del camp.
        /// </summary>
        string Name { get; }
        /// <summary>
        /// Nom visualitzable del camp.
        /// </summary>
        string DisplayName { get; }
        /// <summary>
        /// Visibilitat del camp. Un camp invisible no es tindrà en compte pel ListView.
        /// </summary>
        bool Visible { get; }
    }
    public interface IColumn : IField, ISortedField
    {
    }
    public class Column : IColumn
    {
        //public Column(string Name, string DisplayName)
        //{
        //    this.Name = Name;
        //    this.DisplayName = DisplayName;
        //    this.Visible = true;
        //    this.Direction = ListSortDirection.Ascending;
        //    this.Order = null;
        //}
        /// <summary>
        /// Nom del camp.
        /// </summary>
        public string Name { get; set; }
        /// <summary>
        /// Nom visualitzable del camp.
        /// </summary>
        public string DisplayName { get; set; }
        /// <summary>
        /// Visibilitat del camp. Un camp invisible no es tindrà en compte pel ListView.
        /// </summary>
        public bool Visible { get; set; }
        /// <summary>
        /// Direcció d'ordenació del camp: ascendent o descendent.
        /// </summary>
        public ListSortDirection Direction { get; set; }
        /// <summary>
        /// Ordre del camp a la llista ordenada de camps.
        /// </summary>
        public int? Order { get; set; }
    }
    public interface IListView : IPagedListView, ISortedListView
    {
    }
    public interface IQueryOptions : IPagedQueryOption, ISortedQueryOption
    {
    }
    public interface IListViewModel<T> : IListView, IQueryOptions
    {
        /// <summary>
        /// Llista d'elements amb permisos exclusius de lectura.
        /// </summary>
        IReadOnlyList<T> Items { get; set; }
    }
    public class ListViewModel<T> : IListViewModel<T>, IListViewPaginator, IListViewNavigator
    {
        public static readonly int[] DefaultPageSizeOptions = { 5, 10, 25, 50, 100 };
        public const int DefaultMaxRange = 5;
        public ListViewModel()
        {
            this.Fields = new List<IColumn>();
            this.Items = new List<T>();
            this.PageSizeOptions = DefaultPageSizeOptions;
            this.MaxRange = DefaultMaxRange;
        }

        #region IListViewModel

        #region IQueryOptions

        #region IPagedQueryOptions

        public int Top
        {
            get
            {
                return this.PageSize;
            }
        }
        public int Skip
        {
            get
            {
                return this.PageNumber * this.PageSize;
            }
        }

        #endregion

        #region ISortedQueryOptions

        public SortedList<int, IColumn> SortedFields
        {
            get { throw new System.NotImplementedException(); }
        }

        #endregion

        #endregion

        #region IListView

        /// <summary>
        /// Llista d'elements amb permisos exclusius de lectura.
        /// </summary>
        public IReadOnlyList<T> Items { get; set; }
        /// <summary>
        /// Llista de camps.
        /// </summary>
        public IList<IColumn> Fields { get; set; }

        #endregion

        #region IPagedListView

        public int TotalItemCount { get; set; }
        public int PageSize { get; set; }
        public int PageNumber { get; set; }
        public int PageCount
        {
            get
            {
                return ListViewHelper.PageCount(PageSize, TotalItemCount);
            }
        }

        #endregion

        #endregion

        #region IListViewPaginator

        /// <summary>
        /// Opcions amb el nombre d'elements per pàgina.
        /// </summary>
        public int[] PageSizeOptions { get; set; }
        /// <summary>
        /// Indicador de pàgines posteriors (automàtic).
        /// </summary>
        public bool HasNextPage
        {
            get
            {
                return this.PageNumber < this.PageCount;
            }
        }
        /// <summary>
        /// Indicador de pàgines anteriors (automàtic).
        /// </summary>
        public bool HasPreviousPage
        {
            get
            {
                return this.PageNumber > 1;
            }
        }
        /// <summary>
        /// Indicador de primera pàgina (automàtic).
        /// </summary>
        public bool IsFirstPage
        {
            get
            {
                return this.PageNumber <= 1;
            }
        }
        /// <summary>
        /// Indicador d'última pàgina (automàtic).
        /// </summary>
        public bool IsLastPage
        {
            get
            {
                return this.PageNumber >= this.PageCount;
            }
        }
        /// <summary>
        /// Indicador de segona pàgina (automàtic).
        /// </summary>
        public bool IsSecondPage
        {
            get
            {
                return this.PageNumber == 2;
            }
        }
        /// <summary>
        /// Indicador de penúltima pàgina (automàtic).
        /// </summary>
        public bool IsPenultimatePage
        {
            get
            {
                return this.PageNumber == this.PageCount - 1;
            }
        }
        /// <summary>
        /// Índex del primer element de la pàgina actual (automàtic).
        /// </summary>
        public int? FirstItem
        {
            get
            {
                return ListViewHelper.FirstItemOnPage(this.PageNumber,
                    this.PageSize, this.TotalItemCount);
            }
        }
        /// <summary>
        /// Índex de l'últim element de la pàgina actual (automàtic).
        /// </summary>
        public int? LastItem
        {
            get
            {
                return ListViewHelper.LastItemOnPage(this.PageNumber,
                    this.PageSize, this.TotalItemCount);
            }
        }

        #endregion

        #region IListViewNavigator

        /// <summary>
        /// Nombre màxim de pàgines del rang seleccionable on pertany la pàgina actual.
        /// </summary>
        public int MaxRange { get; set; }
        /// <summary>
        /// Array d'enters amb l'interval de pàgines del navegador.
        /// </summary>
        public int[] Range
        {
            get
            {
                return ListViewHelper.RangeNavigator(this.PageNumber,
                    this.PageCount, this.MaxRange);
            }
        }
        /// <summary>
        /// Pàgina de navegació a l'esquerra o nul si no n'hi ha.
        /// </summary>
        public int? LeftNavigator
        {
            get
            {
                return ListViewHelper.LeftNavigator(this.Range);
            }
        }
        /// <summary>
        /// Pàgina de navegació a la dreta o nul si no n'hi ha.
        /// </summary>
        public int? RightNavigator
        {
            get
            {
                return ListViewHelper.RightNavigator(this.Range, this.PageCount);
            }
        }

        #endregion

    }
}
